/**
 * @SchiebePuzzleGUI.java     06.07.2010
 * build: 06.07.2010
 * Copyright 2010 yourwebs.de. All rights reserved
 * @author J&ouml;rg Czeschla
 *
 * This file is a part of SchiebePuzzle
 * Test is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * SchiebePuzzle is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License along with Test;
 * if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA. 
 */
package de.yourwebs.schiebepuzzle;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.Rectangle;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.JFrame;
import javax.swing.JPanel;

/**
 * Displays a kind of shiftpuzzle based on several square-elements that are
 * arranged as a big square. The play-mode can be changed from pushing the
 * elements for only one square to jumping over several squares. The
 * display-mode can be changed from displaying numbers to displaying a sliced
 * picture.
 */
public class SchiebePuzzleGUI extends JFrame {

	private static final long serialVersionUID = 1L;

	// elements per row/col
	protected static int elements = 4;
	protected int rows = elements;
	protected int cols = elements;

	// colors
	private final Color bgColor = new Color(210, 210, 225);
	private final Color emptyColor = new Color(255, 255, 255);

	// Number of shuffles. The arrays are to be shuffled to place the elements
	// in a random manner.
	// specifies how offen the arrays are shuffled. Each shuffling changes the
	// placees of all the
	// neighbour-elements for the whole array.
	private static final int shuffle = 5;

	// Game mode: true if only the elements next to the empty element can be
	// moved, otherwise false
	protected boolean mode = true;

	// Display mode: true if images are to be shown, otherwise show numbers
	protected boolean display = true;

	// image-path
	private String imgPath = "img/002.jpg";

	// size of the playground-panel
	protected final int borderLength = 600;
	protected static int minCols = 3;
	protected static int maxCols = 25;

	// font-size
	public static int fontSize = 20;

	// Element-arrays
	protected SchiebeElement numberElements[] = new SchiebeElement[rows * cols];
	protected SchiebeElement imageElements[] = new SchiebeElement[rows * cols];
	private SchiebeElement emptyElement;
	protected Rectangle clickedRectBnds, emptyRectBnds;

	// GUI-Components
	protected JPanel mainPanel;
	private ButtonPanel buttPanel;
	private PuzzleModel puzzleModel;

	public SchiebePuzzleGUI() {
		
		this.setLayout(new BorderLayout());

		// playground
		mainPanel = new JPanel();
		mainPanel.setLayout(new GridLayout(rows, cols));
		mainPanel.setSize(borderLength, borderLength);
		mainPanel.setPreferredSize(new Dimension(borderLength, borderLength));

		// managing images in the model
		puzzleModel = new PuzzleModel(mainPanel);
		puzzleModel.setImgPath(imgPath, true);
		puzzleModel.setAktBufferedImage();
		
		// buttons
		buttPanel = new ButtonPanel(this);
		buttPanel.initButtons();

		// create all the elements
		createElements();
		if (display) {
			puzzleModel.setImage(imageElements);
			shuffleElements(imageElements);
			arrangeElements(imageElements);
		} else {
			puzzleModel.setImage(numberElements);
			shuffleElements(numberElements);
			arrangeElements(numberElements);
		}

		
		this.add(mainPanel, BorderLayout.CENTER);
		this.add(buttPanel, BorderLayout.SOUTH);

		this.setTitle("Schiebe-Puzzle");
		this.setIconImage(puzzleModel.loadImage("img/puzzle_512.png"));
		this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		this.pack();
		this.setResizable(false);
		this.setLocationRelativeTo(null);
		this.setVisible(true);
	} // Konstruktor

	protected void rearrange() {
		imageElements = new SchiebeElement[rows * cols];
		numberElements = new SchiebeElement[rows * cols];
		mainPanel.removeAll();
		createElements();
		((GridLayout) mainPanel.getLayout()).setRows(rows);
		((GridLayout) mainPanel.getLayout()).setColumns(cols);
		mainPanel.repaint();
		mainPanel.revalidate();

		if (display) {
			puzzleModel.setImage(imageElements);
			shuffleElements(imageElements);
			arrangeElements(imageElements);
		} else {
			shuffleElements(numberElements);
			arrangeElements(numberElements);
		}
	}

	/**
	 * Creates two arrays of elements: 1. image-elements 2. number-elements the
	 * last element of each SchiebeElement-array is an empty element
	 */
	protected void createElements() {
		SchiebeElement sE;
		for (int i = 0; i < rows * cols - 1; i++) {
			// number-elements
			sE = new NumberElement(mainPanel, i + 1);
			sE.setBackground(bgColor);
			sE.addMouseListener(new PuzzleMouseListener(this));
			numberElements[i] = sE;

			// image-elements
			sE = new ImageElement(mainPanel);
			sE.setBackground(bgColor);
			sE.addMouseListener(new PuzzleMouseListener(this));
			imageElements[i] = sE;
		}
		// create an empty element and add it to the array's end
		emptyElement = new SchiebeElement(mainPanel);
		emptyElement.setBackground(emptyColor);
		numberElements[numberElements.length - 1] = emptyElement;
		imageElements[imageElements.length - 1] = emptyElement;
	}

	/**
	 * Selects two elements randomly and switches their position in the
	 * element's array. Not every element must be changed but can be changed
	 * several times
	 */
	protected void shuffleElements(SchiebeElement[] sE) {
		int rand1, rand2;
		for (int i = 0; i < sE.length * shuffle; i++) {
			rand1 = (int) (Math.random() * sE.length) + 1;
			rand2 = (int) (Math.random() * sE.length) + 1;

			// switch two elements randomly
			if (rand1 >= sE.length)
				rand1 = 0;
			if (rand1 < 0)
				rand1 = sE.length - 1;
			if (rand2 >= sE.length)
				rand2 = 0;
			if (rand2 < 0)
				rand2 = sE.length - 1;

			// switching
			SchiebeElement temp = sE[rand1];
			sE[rand1] = sE[rand2];
			sE[rand2] = temp;
		}
	}

	/**
	 * Adds all elements of the given array to the panel
	 * 
	 * @param se
	 *            Array to be added
	 */
	protected void arrangeElements(SchiebeElement[] se) {
		for (int i = 0; i < se.length; i++) {
			mainPanel.add(se[i]);
		}
	}

	/**
	 * 
	 * @return the empty element
	 */
	public SchiebeElement getEmptyElement() {
		return emptyElement;
	}
	
	public int getBorderLength(){
		return borderLength;
	}
	
	public JPanel getPanel(){
		return mainPanel;
	}
	
	protected PuzzleModel getPuzzleModel(){
		return puzzleModel;
	}

	/**
	 * checks whether an element is a proximat neighbour of the empty element
	 * 
	 * @return true if the asked element is a neighbour otherwise false
	 */
	public boolean isNeighbour() {
		// wenn jedes Element bewegt werden darf, muss
		// der Nachbar nicht ermittelt werden
		if (!mode)
			return true;
		// Position des leeren Rechtecks
		int eRPosX = emptyRectBnds.x;
		int eRPosY = emptyRectBnds.y;
		// Position des angeklickten Rechtecks
		int cRPosX = clickedRectBnds.x;
		int cRPosY = clickedRectBnds.y;
		// Positionen festlegen
		int rectSize = mainPanel.getWidth() / elements;
		boolean o = eRPosX == cRPosX && (eRPosY - rectSize) == cRPosY;
		boolean u = eRPosX == cRPosX && (eRPosY + rectSize) == cRPosY;
		boolean l = (eRPosX - rectSize) == cRPosX && eRPosY == cRPosY;
		boolean r = (eRPosX + rectSize) == cRPosX && eRPosY == cRPosY;
		return o || u || l || r;
	}

	/**
	 * changes the position of an element with that of the empty element by
	 * ascertaining its bounds
	 * 
	 * @param e
	 */
	public void switchElem(MouseEvent e) {
		// only switch neighbours
		if (isNeighbour()) {
			Rectangle tmpBnds = clickedRectBnds;
			clickedRectBnds = emptyRectBnds;
			emptyRectBnds = tmpBnds;

			getEmptyElement().setBounds(emptyRectBnds);
			((SchiebeElement) e.getSource()).setBounds(clickedRectBnds);
		}
	}

	/**
	 * sets the path of an external loaded image, registers it at the model, and
	 * display it
	 * 
	 * @param imgPath
	 *            String path to the image
	 */
	public void setExtImg(String imgPath) {
		puzzleModel.setImgPath(imgPath, false);
		puzzleModel.setAktBufferedImage();
		rearrange();
		mainPanel.revalidate();
	}
}

class PuzzleMouseListener extends MouseAdapter {
	
	SchiebePuzzleGUI gui;
	
	public PuzzleMouseListener(SchiebePuzzleGUI gui){
		this.gui = gui
;	}
	public void mouseReleased(MouseEvent e) {
		// getting the bounds of the empty element and its associate partner
		gui.emptyRectBnds = gui.getEmptyElement().getBounds();
		gui.clickedRectBnds = ((SchiebeElement) e.getSource()).getBounds();
		gui.switchElem(e);
	}
}
